from typing import Union

import pandas as pd
import numpy as np
from numpy.typing import ArrayLike
from icecream import ic

from trimes.base import (
    create_pandas_series_or_frame_with_same_columns_and_index,
    resample,
)
import trimes.signal_processing


def get_fourier_coef_rolling(
    ts: pd.DataFrame | pd.Series,
    samples_per_window: int,
    time_windows: None | ArrayLike = None,
) -> tuple[ArrayLike, ArrayLike]:
    """Get rolling fourier coefficients for each sample of a time series.

    Args:
        ts (pd.DataFrame | pd.Series): time series
        samples_per_window (int): number of samples per window (e.g. number of samples of one period)
        time_windows (None | ArrayLike, optional): If specified, the time period of the rolling windows is adapted. Same length as rows in 'ts'. Defaults to None.

    Returns:
        tuple[ArrayLike, ArrayLike]: Fourier coefficients (real and imaginary) for each sample of the time series (values of first window are extended)
    """
    ts_ndim = ts.ndim
    if ts_ndim == 1:  # work with 2D array
        num_col = 1
        ts = ts.to_frame()
    else:
        num_col = ts.shape[1]

    fourier_coef_real = np.empty_like(ts.to_numpy()).reshape(ts.shape[0], -1)
    fourier_coef_imag = np.empty_like(ts.to_numpy()).reshape(ts.shape[0], -1)

    time = ts.index.values
    if time_windows is None:
        index_first_window = samples_per_window
        get_samples_for_window = lambda idx, col: ts.iloc[
            idx - samples_per_window : idx, col
        ].to_numpy()
    else:
        index_first_window = (
            np.argmax(time - time[0] > time_windows) + 1
        )  # first window where time_window fits in ts
        get_samples_for_window = lambda idx, col: resample(
            ts.iloc[:, col],
            np.linspace(time[idx] - time_windows[idx], time[idx], samples_per_window),
        ).to_numpy()
    for col in range(num_col):
        fc_real_col = np.empty_like(ts.iloc[index_first_window:, col].to_numpy())
        fc_imag_col = np.empty_like(ts.iloc[index_first_window:, col].to_numpy())
        for idx in range(index_first_window, ts.shape[0]):
            samples = get_samples_for_window(idx, col)
            fc_real_col[idx - index_first_window] = np.sqrt(2) * get_fourier_coef_real(
                samples
            )
            fc_imag_col[idx - index_first_window] = np.sqrt(2) * get_fourier_coef_imag(
                samples
            )
        fourier_coef_real[:, col] = trimes.signal_processing.extend_np(
            fc_real_col, index_first_window, "wrap"
        )
        fourier_coef_imag[:, col] = trimes.signal_processing.extend_np(
            fc_imag_col, index_first_window, "wrap"
        )
    if ts_ndim == 1:
        fourier_coef_real = fourier_coef_real.squeeze()
        fourier_coef_imag = fourier_coef_imag.squeeze()
    return fourier_coef_real, fourier_coef_imag


def get_fourier_coef_real(x: ArrayLike, k: int = 1, angle: float = 0.0) -> float:
    """Get the real part of the Fourier coefficient.

    Args:
        x (ArrayLike): samples
        k (int, optional): harmonic number. Defaults to 1.
        angle (float, optional): starting angle of cosine. Defaults to 0.0.

    Returns:
        float: Real part of Fourier coefficient
    """
    cos = np.cos(
        np.linspace(angle, k * 2 * np.pi + angle - 2 * np.pi / len(x), k * len(x))
    )
    return np.mean(cos * x)


def get_fourier_coef_imag(x: ArrayLike, k: int = 1, angle: float = 0.0) -> float:
    """Get the imaginary part of the Fourier coefficient.

    Args:
        x (ArrayLike): samples
        k (int, optional): harmonic number. Defaults to 1.
        angle (float, optional): starting angle of sine. Defaults to 0.0.

    Returns:
        float: Imaginary part of Fourier coefficient
    """
    sin = np.sin(
        np.linspace(angle, k * 2 * np.pi + angle - 2 * np.pi / len(x), k * len(x))
    )
    return np.mean(sin * x)



"""Generic methods for signal processing"""

from typing import Union

import pandas as pd
import numpy as np
from numpy.typing import ArrayLike
from scipy.optimize import minimize_scalar
from icecream import ic


from trimes.base import (
    create_pandas_series_or_frame_with_same_columns_and_index,
    resample,
)
import trimes.fourier


def extend_ts(
    ts: Union[pd.DataFrame, pd.Series], n: int, mode: str = "wrap"
) -> Union[pd.DataFrame, pd.Series]:
    """Extend array by 'n' samples in the beginning or end of the array.

    See 'extend_np' for more details.

    Args:
        x (ArrayLike): array to be extended (1D or 2D)
        n (int): number of samples of the extension. Positive extends the array in the beginning, negative extends the array in the end.
        mode (str): 'wrap' simply copies the samples from the first/last n samples of 'x', 'reflect' flips the samples. Defaults to 'wrap'.

    Returns:
        Union[pd.DataFrame, pd.Series]: extended time series
    """
    ts_extended_np = extend_np(ts.to_numpy(), n, mode)
    if n > 0:
        extension = ts.index[0] - ts.index[n] + ts.index[:n]
        index = np.concat((extension, ts.index))
    else:
        extension = ts.index[-1] - ts.index[n - 1] + ts.index[n:]
        index = np.concat((ts.index, extension))
    ts_extended = pd.DataFrame(ts_extended_np, index=index, columns=ts.columns)
    ts_extended.index.name = ts.index.name
    return ts_extended.squeeze()


def extend_np(x: ArrayLike, n: int, mode: str = "wrap") -> ArrayLike:
    """Extend array by 'n' samples in the beginning or end of the array.

    A typical use case is to extend the array by 'n' samples in the beginning of the array, so that the rolling average can be calculated for the first 'n' samples.


    Args:
        x (ArrayLike): array to be extended (1D or 2D)
        n (int): number of samples of the extension. Positive extends the array in the beginning, negative extends the array in the end.
        mode (str): 'wrap' simply copies the samples from the first/last n samples of 'x', 'reflect' flips the samples. Defaults to 'wrap'.

    Raises:
        ValueError: if 'x' is not 1D or 2D.

    Returns:
        ArrayLike: extended array
    """
    ndim = x.ndim
    if ndim == 1:  # work with 2D arrays
        x = x.reshape(-1, 1)
    elif ndim > 2:
        raise ValueError("x must be 1D or 2D array")
    if n > 0:
        extension = x[:n, :]
    else:
        extension = x[n:, :]
    if mode == "wrap":
        pass
    elif mode == "reflect":
        extension = np.flip(extension, axis=0)
    if n > 0:
        x = np.concatenate((extension, x), axis=0)
    else:
        x = np.concatenate((x, extension), axis=0)
    if ndim == 1:
        x = x.reshape(-1)
    return x


def get_angle(x: ArrayLike, **kwargs) -> float:
    """Get angle of periodic time series 'x'.

    TODO: Just use np.angle(fourier_coeff) to get angle.

    Get angle by maximizing (minimizing negative) real part of fourier coefficient.

    Args:
        x (ArrayLike): One period of time series.

    Returns:
        float: angle [rad]
    """
    f = lambda y, args: -trimes.fourier.get_fourier_coef_real(args, angle=y)
    opt_res = minimize_scalar(f, args=(x), bounds=(-np.pi, np.pi), **kwargs)
    return opt_res.x


def average_rolling(
    ts: Union[pd.DataFrame, pd.Series],
    samples_per_window: int,
) -> Union[pd.DataFrame, pd.Series]:
    """Get rolling average of time series 'ts' using 'samples_per_window' samples.

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        samples_per_window (int): number of samples per window

    Returns:
        Union[pd.DataFrame, pd.Series]: Averaged time series (same shape as 'ts', first window is extended)
    """
    ts_average = ts.copy()
    ts_ndim = ts.ndim
    if ts_ndim == 1:
        ts_average = ts_average.to_frame()
        num_col = 1
        ts = ts.to_frame()
    else:
        num_col = ts.shape[1]
    for col in range(
        num_col
    ):  # Cannot vectorize this loop because dtypes of columns can be different in DataFrames
        ts_cumsum = np.cumsum(ts.iloc[:, col].to_numpy())
        avg = (
            ts_cumsum[samples_per_window:] - ts_cumsum[:-samples_per_window]
        ) / samples_per_window
        ts_average.iloc[:, col] = extend_np(avg, samples_per_window, "wrap")
    if ts_ndim == 1:
        ts_average = ts_average.squeeze()
    return ts_average


def average_rolling_variable_window(
    ts: Union[pd.DataFrame, pd.Series],
    time_windows: ArrayLike,
    samples_per_window: int,
) -> Union[pd.DataFrame, pd.Series]:
    """Get rolling average of time series 'ts' using variable time windows that are resampled with 'samples_per_window' number of samples.

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        time_windows (np.array): duration of time windows (same length as rows in 'ts')
        samples_per_window (int): number of samples per window


    Returns:
        Union[pd.DataFrame, pd.Series]: Averaged time series (same shape as 'ts', first window is extended)
    """
    time = ts.index.values
    index_first_window = (
        np.argmax(time - time[0] > time_windows) + 1
    )  # first window where time_window fits in ts
    ts_average = ts.copy()
    ts_ndim = ts.ndim
    if ts_ndim == 1:
        ts_average = ts_average.to_frame()
        num_col = 1
        ts = ts.to_frame()
    else:
        num_col = ts.shape[1]
    for col in range(num_col):
        avg_col = np.empty_like(ts.iloc[index_first_window:, col].to_numpy())
        for idx in range(index_first_window, ts.shape[0]):
            t_samples_window = np.linspace(
                time[idx] - time_windows[idx],
                time[idx],
                samples_per_window,
            )
            samples = resample(ts.iloc[:, col], t_samples_window)
            avg_col[idx - index_first_window] = np.mean(samples)
        ts_average.iloc[:, col] = extend_np(avg_col, index_first_window, "wrap")
    if ts_ndim == 1:
        ts_average = ts_average.squeeze()
    return ts_average



from collections.abc import Callable
from numbers import Number

import numpy as np
from numpy.typing import ArrayLike
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from icecream import ic

from trimes.base import superpose_series, get_index


class PeriodicSignal:
    """Create periodic time series signals."""

    def __init__(
        self,
        t: ArrayLike,
        func: Callable = np.cos,
        f: ArrayLike | tuple = 1.0,
        mag: ArrayLike | tuple = 1.0,
        offset: float = 0.0,
        phi: float = 0.0,
    ) -> None:
        self.t: ArrayLike = t
        "time"
        self.func: Callable = func
        "function to create periodic signals (e.g. np.sin)"
        if isinstance(f, tuple):
            f = np.linspace(f[0], f[1], len(self.t))
        self.f: ArrayLike = f
        "frequency"
        if isinstance(mag, tuple):
            mag = np.linspace(mag[0], mag[1], len(self.t))
        self.mag: ArrayLike = mag
        "magnitude"
        if isinstance(offset, tuple):
            offset = np.linspace(offset[0], offset[1], len(self.t))
        self.offset: ArrayLike = offset
        "y offset"
        self.phi: float = phi
        "initial angle"

    def get_signal(self) -> ArrayLike:
        return self.mag * self.func(self.get_angle_over_time()) + self.offset

    def get_signal_from_angle(self, angle: ArrayLike) -> ArrayLike:
        return self.mag * self.func(angle) + self.offset

    def get_signal_series(self) -> pd.Series:
        series = pd.Series(self.get_signal(), index=self.t)
        series.index.name = "time"
        return series

    def get_signal_series_from_angle(self, angle: ArrayLike) -> pd.Series:
        series = pd.Series(self.get_signal_from_angle(angle), index=self.t)
        series.index.name = "time"
        return series

    def get_signal_n_phases(self, n: int) -> pd.DataFrame:
        return superimpose_and_concat_periodic_signals([[self]], num_phases=n)

    def get_angle_over_time(self) -> ArrayLike:
        """Get angle over time.

        Calculation is dependent on the frequency format (constant scalar number or gradient).
        """
        if isinstance(self.f, Number):
            return 2 * np.pi * self.f * (self.t - self.t[0]) + self.phi
        else:
            frequency_integrated = (
                np.cumsum(2 * np.pi * self.f[1:] * np.diff(self.t)) + self.phi
            )
            return np.array([self.phi, *frequency_integrated])

    def get_signal_attributes_over_time(
        self, attributes: list[str] | None = None
    ) -> pd.DataFrame:
        if attributes is None:
            attributes = ["f", "mag", "offset"]
        attrs_over_time = pd.DataFrame(columns=attributes, index=self.t, dtype=float)
        for attr in attributes:
            attrs_over_time[attr] = self.__getattribute__(attr)
        return attrs_over_time

    def plot(self, **kwargs):
        series = self.get_signal_series()
        series.plot(**kwargs)

    def plot_with_attributes(self, attributes: list[str] | None = None) -> list[Axes]:
        """Plot signal with attributes

        Args:
            attributes (list[str] | None, optional): Atrributes ('f', 'mag', 'offset'). Defaults to None (all).

        Returns:
            list[Axes]: Axes of plot.
        """
        attr = self.get_signal_attributes_over_time(attributes)
        fig, axes = plt.subplots(nrows=attr.shape[1] + 1, ncols=1)
        fig.tight_layout()
        plt.sca(axes[0])
        self.plot()
        plt.title("signal")
        plt.xlabel("")
        for col, ax in enumerate(axes[1:]):
            ax.plot(attr.index, attr.iloc[:, col])
            ax.grid()
            ax.set_title(attr.columns[col])
        return axes


def superimpose_and_concat_periodic_signals(
    signals: list[list[PeriodicSignal]],
    num_phases: int = 1,
    angle_continuation: bool = True,
) -> pd.Series | pd.DataFrame:
    """Concatenate array of periodic signals.

    Args:
        signals (list[list[PeriodicSignal]]): List of lists of PeriodicSignal objects. The PeriodicSignals in each row are superimposed and the resulting column is concatenated.

        num_phases (int, optional): To create multiples phases with angle separation of pi/num_phases. Defaults to 1.

        angle_continuation (bool, optional): If true, jumps in phase angle between signals that are concatenated are avoided (the phase angle PeriodicSignal.phi is adapted).  Defaults to True.

    Returns:
         pd.Series | pd.DataFrame: Series or DataFrame depending on the number of phases.

    Examples:
        Define array of PeriodicSignals. Rows will be superimposed and the resulting column will be concatenated:

        ```
        step_size = 1e-3
        signals = [
            [
                PeriodicSignal(np.arange(0, 0.1, step_size), f=(50, 52)),
                PeriodicSignal(np.arange(0, 0.1, step_size), f=5*50, mag=(0.03, 0.06)),
            ],
            [
                PeriodicSignal(np.arange(0.1, 0.3, step_size), f=50),
                PeriodicSignal(np.arange(0.1, 0.3, step_size), f=7*50, mag=0.06),
            ],
            [
                PeriodicSignal(np.arange(0.3, 0.5, step_size), f=49),
                PeriodicSignal(np.arange(0.3, 0.5, step_size), f=9*50, mag=0.06),
            ],
        ]
        res = concat_periodic_signals(signals, num_phases = 3)
        ```
    """
    phase_signals = [None] * num_phases
    for phase in range(num_phases):
        phase_angle = phase * (-2 * np.pi / num_phases)
        series_to_concat = [None] * np.shape(signals)[0]
        if angle_continuation:
            phi = [signal.phi for signal in signals[0]]
        for n, signals_to_superpose in enumerate(signals):
            signals_superposed = [None] * np.shape(signals)[1]
            for m, signal in enumerate(signals_to_superpose):
                phi_initial = signal.phi
                if angle_continuation:
                    signal.phi = phi[m]
                signal.phi += phase_angle
                angle = signal.get_angle_over_time()
                signals_superposed[m] = signal.get_signal_series_from_angle(angle)
                if angle_continuation:
                    phi[m] = angle[-1] + (angle[-1] - angle[-2]) - phase_angle
                signal.phi = phi_initial
            series_to_concat[n] = superpose_series(signals_superposed)

        phase_signals[phase] = pd.concat(series_to_concat, axis=0)
    if num_phases == 1:
        return phase_signals[0]
    else:
        return pd.concat(phase_signals, axis=1)


def get_attributes_of_superimposed_and_concatenated_signals_over_time(
    signals: list[list[PeriodicSignal]],
    columns: list[int] | None = None,
    attributes: list[str] | None = None,
) -> list[pd.DataFrame]:
    """_summary_

    Args:
        signals (list[list[PeriodicSignal]]): List of lists of PeriodicSignal objects. The PeriodicSignals in each row are superimposed and the resulting column is concatenated.

        columns (list[int] | None, optional): Columns for which attribute are returned. Defaults to None (all columns).

        attributes (list[str] | None, optional): Attributes ('f', 'mag', 'offset'). Defaults to None (all).

    Returns:
        list[pd.DataFrame]: Attributes for columns.
    """
    if columns is None:
        columns = range(0, len(signals[0]))
    attrs_of_columns = [None] * len(columns)
    for n, col in enumerate(columns):
        frames = [pd.DataFrame()] * len(signals)
        for row in range(len(signals)):
            frames[row] = signals[row][col].get_signal_attributes_over_time(attributes)
        attrs_of_columns[n] = pd.concat(frames)
    return attrs_of_columns


def linear_time_series(
    t: ArrayLike, y: ArrayLike, sample_time: float
) -> pd.DataFrame | pd.Series:
    """Get linear time series.

    Args:
        t (ArrayLike): points in time
        y (ArrayLike): y values corresponding to 't'
        sample_time (float): sample time

    Returns:
        pd.DataFrame | pd.Series: time series
    """
    t = make_monotonously_increasing(t, sample_time)
    ts = get_interpolated(t, y, sample_time)
    return ts


def make_monotonously_increasing(t: ArrayLike, sample_time: float) -> ArrayLike:
    """Make time 't' monotonously increasing.

    Two consecutive values n, k in 't', where k <= n, are replaced with n, k+sample_time.

    Args:
        t (ArrayLike): time
        sample_time (float): sample time

    Returns:
        ArrayLike: monotonously increasing time
    """
    t = np.array(t, dtype=float)
    diff_t_is_zero = np.concat([[False], np.diff(t) <= 0])
    t[diff_t_is_zero] = t[diff_t_is_zero] + sample_time
    return t


def get_interpolated(
    t: ArrayLike, y: ArrayLike | list[ArrayLike], sample_time: float
) -> pd.DataFrame | pd.Series:
    """Linear interpolated y (specified at 'sample_time') over 't'.

    Args:
        t (ArrayLike): time

        y (ArrayLike | list[ArrayLike]): y values (length of first dimension matches 'sample_time') or list of y values to get several results.

        sample_time (float): sample time (length of first dimension matches 'y')

    Returns:
        pd.DataFrame | pd.Series: interpolated values
    """
    index = np.arange(t[0], t[-1] + sample_time, sample_time)
    if not isinstance(y, list):
        return_series = True
        y = [y]
    else:
        return_series = False
    ts = pd.DataFrame(np.nan, index=index, columns=range(len(y)))
    ts.iloc[get_index(ts, t), :] = np.transpose(y)
    ts.interpolate(inplace=True)
    ts.index.name = "time"
    if return_series:
        ts = ts.squeeze()
    return ts


def mirror_y(ts: pd.Series, y: float, inplace=False) -> pd.DataFrame | pd.Series:
    """Mirror 'ts' at 'y'.

    Args:
        ts (pd.Series): time series

        y (float): y (value of horizontal line for mirroring)

        inplace (bool, optional): If True the mirrored signal is added to 'ts' and a DataFrame is returned. Defaults to False (only mirrored signal series is returned).

    Returns:
        pd.DataFrame | pd.Series: Mirrored series or frame
    """
    mirrored_y = 2 * y - ts
    if inplace:
        ts = ts.to_frame()
        ts[1] = mirrored_y
        return ts
    else:
        return mirrored_y


def create_discrete_steps_for_interpolated_time_series(
    t: np.array, val: np.array, delta_t: float = 1e-6, name="0"
) -> pd.Series:
    """Create a time series that approximates discrete value steps with samples that are close in time.

    E.g. a step in 'val' at time 't' from 0 to 1 is approximated with two samples at 't - delta_t' ('val=0') and 't' ('val=1'), where 'delta_t' is a small value (e.g. 1e-6).

    Args:
        t (np.array): time of discrete steps
        val (np.array): New value at discrete steps 't'
        delta_t (float, optional): time difference at discrete steps. Defaults to 1e-6.
        name (str, optional): label in pandas series. Defaults to "0".

    Returns:
        pd.Series: time series with discrete steps
    """
    num_time_steps = len(t)
    time_ticks_for_interpolation = np.empty(num_time_steps * 2 - 1, dtype=float)
    values_for_interpolation = np.empty(num_time_steps * 2 - 1, dtype=float)
    for idx in range(num_time_steps - 1):
        idx2 = idx * 2
        values_for_interpolation[idx2 : idx2 + 2] = [val[idx], val[idx]]
        time_ticks_for_interpolation[idx2 : idx2 + 2] = [
            t[idx],
            t[idx + 1] - delta_t,
        ]
    time_ticks_for_interpolation[-1] = t[-1]
    values_for_interpolation[-1] = val[-1]
    ts = pd.Series(values_for_interpolation, index=time_ticks_for_interpolation)
    ts.name = name
    ts.index.name = "time"
    return ts



"""Convert from .ipynb to .qmd (or reverse)
"""

import glob, os
import subprocess

ipynb2qmd = True
subdirectory: str | None = "./docs/tutorials"


if __name__ == "__main__":
    if subdirectory:
        os.chdir(subdirectory)
    print(f"Current working directory: {os.getcwd()}")
    if ipynb2qmd:
        file_type = "*.ipynb"
    else:
        file_type = "*.qmd"
    files = glob.glob(file_type)
    if not files:
        print(f"No {file_type} found in {os.getcwd()}.")
    for file in glob.glob(file_type):
        subprocess.run(f"quarto convert {file}".split(" "))



import numpy as np


def pt1(x, Tf, sample_time):
    """Implement a first-order low-pass filter.

    The input data is x, the filter's cutoff frequency is omega_c
    [rad/s] and the sample time is T [s].  The output is y.
    """
    y = np.empty(x.shape[0], dtype=float)
    y[0] = x[0]
    for k in np.arange(1, x.shape[0]):
        y[k] = y[k - 1] + (x[k] - y[k - 1]) / Tf * sample_time
    return y



"""
Calculate quantities of electric power systems.
"""

import numpy as np
from numpy.typing import ArrayLike
import pandas as pd
from icecream import ic

from trimes.root_mean_square import rms_rolling, rms_rolling_variable_window
from trimes.base import resample
from trimes.fourier import (
    get_fourier_coef_rolling,
)
from trimes.transforms import abc_2_symmetrical_components
from trimes.signal_processing import (
    average_rolling_variable_window,
    average_rolling,
)


def get_apparent_power_symmetrical_components(
    u: ArrayLike, i: ArrayLike, factor: float = 3
) -> complex:
    """Get apparent power from symmetrical components of voltages and currents.

    Args:
        u (ArrayLike): Voltage pos., neg. and zero sequence components
        i (ArrayLike): Current pos., neg. and zero sequence components
        factor (float, optional): Multiplication factor. Defaults to 3.

    Returns:
        ArrayLike: Apparent power
    """
    return factor * u * np.conj(i)


def get_active_power_symmetrical_components(
    u: ArrayLike, i: ArrayLike, factor: float = 3
):
    return factor * (np.dot(np.real(u), np.real(i)) + np.dot(np.imag(u), np.imag(i)))


def get_reactive_power_symmetrical_components(
    u: ArrayLike, i: ArrayLike, factor: float = 3
):
    return factor * (-np.dot(np.real(u), np.imag(i)) + np.dot(np.imag(u), np.real(i)))


def get_apparent_power_using_symmetrical_components_and_fourier_coefficients(
    u_abc: pd.DataFrame,
    i_abc: pd.DataFrame,
    samples_per_window: int,
    time_windows: ArrayLike | None = None,
    return_symmetrical_components: bool = False,
) -> ArrayLike:
    """Get apparent power of symmetrical components

    Steps:
    - calculate fourier coefficients
    - calculate symmetrical components of v and i
    - calculate apparent power

    Args:
        u_abc (pd.DataFrame): three-phase voltages

        i_abc (pd.DataFrame): three-phase currents

        samples_per_window (int): Samples per window (period) to calculate fourier coefficients

        time_windows (ArrayLike | None, optional): Time windows for fourier coefficient calculation (voltages and currents are resampled for each time window). Defaults to None (no resampling).

        return_symmetrical_components (bool, optional): If True, the sym. comp. of voltages and currents are returned. Defaults to False.

    Returns:
        ArrayLike: Apparent power of sym. components (+ sym. comp. of v and i if 'return_symmetrical_components' is True)
    """
    n = 5
    if time_windows is not None:
        u_fourier_coef_real, u_fourier_coef_imag = get_fourier_coef_rolling(
            u_abc,
            samples_per_window,
            time_windows=time_windows,
        )

        i_fourier_coef_real, i_fourier_coef_imag = get_fourier_coef_rolling(
            i_abc,
            samples_per_window,
            time_windows=time_windows,
        )
    else:
        u_fourier_coef_real, u_fourier_coef_imag = get_fourier_coef_rolling(
            u_abc,
            samples_per_window,
        )
        i_fourier_coef_real, i_fourier_coef_imag = get_fourier_coef_rolling(
            i_abc,
            samples_per_window,
        )

    u_phasors_abc = u_fourier_coef_real + u_fourier_coef_imag * 1j
    i_phasors_abc = i_fourier_coef_real + i_fourier_coef_imag * 1j

    u_sym_comp = np.apply_along_axis(abc_2_symmetrical_components, 1, u_phasors_abc)
    i_sym_comp = np.apply_along_axis(abc_2_symmetrical_components, 1, i_phasors_abc)
    if not return_symmetrical_components:
        return get_apparent_power_symmetrical_components(u_sym_comp, i_sym_comp)
    else:
        return (
            get_apparent_power_symmetrical_components(u_sym_comp, i_sym_comp),
            u_sym_comp,
            i_sym_comp,
        )


def get_instantaneous_active_power_3ph(u_abc: ArrayLike, i_abc: ArrayLike) -> ArrayLike:
    """u_a*i_a + u_b*i_b + u_c*i_c

    Args:
        u_abc (ArrayLike): three-phase voltages
        i_abc (ArrayLike): three-phase currents

    Returns:
        ArrayLike: _description_
    """
    return np.sum(u_abc * i_abc, axis=1)


def get_instantaneous_reactive_power_3ph(
    u_abc: ArrayLike, i_abc: ArrayLike
) -> ArrayLike:
    """From PowerFactory's 'Technical Reference - Common Result Variables for Terminals and Elements (EMT Simulation)'.

    Args:
        u_abc (ArrayLike): three-phase voltages
        i_abc (ArrayLike): three-phase currents

    Returns:
        ArrayLike: instantaneous reactive power.
    """
    return (
        1
        / np.sqrt(3)
        * (
            (u_abc[:, 1] - u_abc[:, 2]) * i_abc[:, 0]
            + (u_abc[:, 2] - u_abc[:, 0]) * i_abc[:, 1]
            + (u_abc[:, 0] - u_abc[:, 1]) * i_abc[:, 2]
        )
    )


def get_instantaneous_reactive_power_using_phase_shift_3ph(
    u_abc: pd.DataFrame,
    i_abc: pd.DataFrame,
    periods: float | ArrayLike,
) -> ArrayLike:
    """
    Calculate instantaneous reactive power using 90 degree phase shift of the current.

    From 'The Measurement of Reactive Energy in Polluted Distribution Power Systems: An Analysis of the Performance of Commercial Static Meters'

    Args:
        u_abc (pd.DataFrame): three-phase voltages
        i_abc (pd.DataFrame): three-phase currents
        periods (ArrayLike): Duration of one period in seconds or array of durations (same length as u_abc/i_abc) if periods are not constant.

    Returns:
        ArrayLike: instantaneous reactive power.
    """
    t_i_abc = i_abc.index.values + periods / 4
    i_abc_shifted = resample(i_abc, t_i_abc)
    return pd.Series(
        np.sum(u_abc.to_numpy()[-len(t_i_abc) :] * i_abc_shifted, axis=1), index=t_i_abc
    )


def get_instantaneous_magnitude_3ph(abc: ArrayLike) -> ArrayLike:
    """Calculate the instantaneous magnitude of a three-phase signal.

        According to “IEEE Std 1459-2010 Standard Definitions for the Measurement
    of Electric Power Quantities Under Sinusoidal, Nonsinusoidal, Balanced,
    or Unbalanced Conditions,” 2010.

        Args:
            abc (ArrayLike): three-phase signal

        Returns:
            ArrayLike: magnitude
    """
    return np.sqrt(
        1
        / 9
        * (
            (abc[:, 0] - abc[:, 1]) ** 2
            + (abc[:, 1] - abc[:, 2]) ** 2
            + (abc[:, 2] - abc[:, 0]) ** 2
        )
    )


def get_active_power_average_rolling_3ph(
    u_abc: ArrayLike, i_abc: ArrayLike, samples_per_window: int
) -> pd.DataFrame | pd.Series:
    """Calculate the active power using a rolling window of constant number of samples.

    Args:
        u_abc (ArrayLike): three-phase voltages
        i_abc (ArrayLike): three-phase currents
        samples_per_window (int): number of samples per window (e.g. number of samples per period)

    Returns:
        pd.DataFrame | pd.Series: active power
    """
    p_df = pd.DataFrame(get_instantaneous_active_power_3ph(u_abc, i_abc))
    return average_rolling(p_df, samples_per_window=samples_per_window)


def get_active_power_average_rolling_variable_window_3ph(
    u_abc: pd.DataFrame,
    i_abc: pd.DataFrame,
    time_windows: ArrayLike,
    samples_per_window: int,
) -> pd.DataFrame | pd.Series:
    """Calculate the active power using a rolling window of variable duration.

    Voltages and currents are resampled for each time window.

    Args:
        u_abc (ArrayLike): three-phase voltages
        i_abc (ArrayLike): three-phase currents
        time_windows (ArrayLike): time periods for the rolling window (same length as u_abc/i_abc)
        samples_per_window (int): number of samples per window

    Returns:
        pd.DataFrame | pd.Series: active power
    """
    p_df = pd.DataFrame(
        get_instantaneous_active_power_3ph(u_abc.to_numpy(), i_abc.to_numpy()),
        index=u_abc.index,
    )
    return average_rolling_variable_window(
        p_df, time_windows, samples_per_window=samples_per_window
    )


def get_reactive_power_average_rolling_3ph(
    u_abc: ArrayLike, i_abc: ArrayLike, samples_per_window: int
) -> pd.DataFrame | pd.Series:
    """Calculate the reactive power using a rolling window of constant number of samples.

    Args:
        u_abc (ArrayLike): three-phase voltages
        i_abc (ArrayLike): three-phase currents
        samples_per_window (int): number of samples per window (e.g. number of samples per period)

    Returns:
        pd.DataFrame | pd.Series: reactive power
    """
    p_df = pd.DataFrame(get_instantaneous_reactive_power_3ph(u_abc, i_abc))
    return average_rolling(p_df, samples_per_window=samples_per_window)


def get_reactive_power_average_rolling_variable_window_3ph(
    u_abc: pd.DataFrame,
    i_abc: pd.DataFrame,
    time_windows: np.array,
    samples_per_window: int,
) -> pd.DataFrame | pd.Series:
    """Calculate the reactive power using a rolling window of variable duration.

    Voltages and currents are resampled for each time window.

    Args:
        u_abc (ArrayLike): three-phase voltages
        i_abc (ArrayLike): three-phase currents
        time_windows (ArrayLike): time periods for the rolling window (same length as u_abc/i_abc)
        samples_per_window (int): number of samples per window

    Returns:
        pd.DataFrame | pd.Series: reactive power
    """
    p_df = pd.DataFrame(
        get_instantaneous_reactive_power_3ph(u_abc.to_numpy(), i_abc.to_numpy()),
        index=u_abc.index,
    )
    return average_rolling_variable_window(
        p_df, time_windows, samples_per_window=samples_per_window
    )



import control
import pandas as pd
import matplotlib.pyplot as plt
from icecream import ic

from trimes.plots import (
    add_point_to_plot,
    add_vertical_line_to_plot,
)


def step_info_series(
    ts: pd.Series, yinitial: float | None = None, yfinal: float | None = None, **kwargs
) -> dict:
    """Get information on step response. Interface to 'control' package's function 'step_info'.

    See python-control.readthedocs.io/en/latest/generated/control.step_info.html

    Args:
        ts (pd.Series): time series

        yinitial (float | None, optional): Initial value before step. Defaults to None (first value in 'ts' is used).

        yfinal (float | None, optional): Final steady state value after step. Defaults to None (last value in 'ts' is used).

        kwargs: kwargs for control.step_info

    Returns:
        dict: step information
    """
    ts_np = ts.to_numpy()
    if yinitial is None:
        yinitial = ts_np[0]
    ts_np = ts_np - yinitial
    return control.step_info(ts_np, ts.index.values, yfinal=yfinal, **kwargs)


def plot_step_info(
    ts: pd.Series,
    step_info: dict,
    yinitial: float | None = None,
    alpha=0.7,
    linestyle="--",
) -> None:
    """Plot step information.

    Args:
        ts (pd.Series): time series

        step_info (dict): corresponding step info

        yinitial (float | None, optional): Initial value before step. Defaults to None (first value in 'ts' is used).

        alpha (float, optional): plot argument. Defaults to 0.7.

        linestyle (str, optional): plot argument. Defaults to "--".
    """
    if yinitial is None:
        yinitial = ts.iloc[0]
    y_final = ts.iloc[-1]

    ax = plt.gca()
    plt_args = {
        "alpha": alpha,
        "linestyle": linestyle,
    }
    for info, val in step_info.items():
        if info in ["SettlingTime", "PeakTime"]:
            add_vertical_line_to_plot(
                val,
                label=info,
                color=ax._get_lines.get_next_color(),
                **plt_args,
            )
        elif info in ["SettlingMin", "SettlingMax", "SteadyStateValue"]:
            val += yinitial
            plt.axhline(
                y=val,  # vertical line
                label=info,
                color=ax._get_lines.get_next_color(),
                **plt_args,
            )
        elif info in ["Overshoot"]:
            if val != 0:
                val = y_final + val / 100 * (y_final - yinitial)
                plt.axhline(
                    y=val,  # vertical line
                    label=info,
                    color=ax._get_lines.get_next_color(),
                    **plt_args,
                )
        elif info in ["Undershoot"]:
            if val != 0:
                val = yinitial - val / 100 * (y_final - yinitial)
                print(val)
                plt.axhline(
                    y=val,  # vertical line
                    label=info,
                    color=ax._get_lines.get_next_color(),
                    **plt_args,
                )
        elif info in ["Peak"]:
            t = step_info.get("PeakTime")
            if t:
                val += yinitial
                add_point_to_plot(t, val, va="bottom", label=info, color="red")
            else:
                raise Exception("'PeakTime' must be available to plot 'Peak'")



from typing import Union

import pandas as pd
import numpy as np
from icecream import ic

from trimes.base import (
    create_pandas_series_or_frame_with_same_columns_and_index,
    resample,
)
from trimes.signal_processing import extend_np


def rms_rolling(
    ts: Union[pd.DataFrame, pd.Series],
    samples_per_window: int,
    normalize_magnitude: bool = False,
) -> Union[pd.DataFrame, pd.Series]:
    """Calculate the root mean square (RMS) of a time series using a rolling window.

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        samples_per_window (int): number of samples per window (e.g. for one period)
        normalize_magnitude (bool, optional): If true, results are multiplied by sqrt(2) (e.g. for per unit values). Defaults to False.

    Returns:
        Union[pd.DataFrame, pd.Series]: time series with RMS values
    """
    normalization_factor = np.sqrt(2) if normalize_magnitude else 1
    ts_squared = np.cumsum(np.power(ts.to_numpy(), 2), axis=0)
    func_rms = (
        lambda x: np.sqrt(
            (x[samples_per_window:] - x[:-samples_per_window]) / samples_per_window
        )
        * normalization_factor
    )
    rms = np.apply_along_axis(func_rms, 0, ts_squared)
    rms = extend_np(rms, samples_per_window, "wrap")
    return create_pandas_series_or_frame_with_same_columns_and_index(rms, ts)


def rms_rolling_variable_window(
    ts: Union[pd.DataFrame, pd.Series],
    time_windows: np.array,
    samples_per_window: int,
    normalize_magnitude: bool = False,
) -> Union[pd.DataFrame, pd.Series]:
    """Calculate the root mean square (RMS) of a time series using a rolling window with variable length.
    The length of the window is defined by the time_windows parameter. The time_windows parameter must be the

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        time_windows (np.array): Duration of windows (e.g. periods for variable frequencies). Same length as the number of rows in the time series.
        samples_per_window (int): Number of samples per window.
        normalize_magnitude (bool, optional): If true, results are multiplied by sqrt(2) (e.g. for per unit values). Defaults to False.

    Returns:
        Union[pd.DataFrame, pd.Series]: RMS values of the time series
    """
    normalization_factor = np.sqrt(2) if normalize_magnitude else 1
    time = ts.index.values
    index_first_window = (
        np.argmax(time - time[0] < time_windows) + 1
    )  # first window where time_window fits in ts
    rms = np.empty_like(ts.to_numpy()).reshape(ts.shape[0], -1)[index_first_window:, :]
    for idx in range(ts.shape[0] - index_first_window):
        t_samples_window = np.linspace(
            time[idx] - time_windows[idx], time[idx], samples_per_window
        )
        samples = resample(ts, t_samples_window)
        rms[idx, :] = (
            np.sqrt(np.sum(np.power(samples, 2), axis=0) / samples_per_window)
            * normalization_factor
        )
    rms = extend_np(rms, index_first_window, "wrap")
    return create_pandas_series_or_frame_with_same_columns_and_index(rms.squeeze(), ts)


def rms_min_max(
    ts: pd.DataFrame,
    filter_time_constant: float,
) -> pd.Series:
    """Get apprmation of RMS value of a 3-phase signal using difference between maxima and minima of the phases and low pass filtering to reduce the ripple.

    According to www.pscad.com/webhelp/EMTDC_Tools_Library/Meters/vm3ph2

    Args:
        ts (pd.DataFrame): time series (3 phases)
        filter_time_constant (float): low pass filter time constant in seconds, e.g. half a period of the fundamental frequency

    Returns:
        pd.Series: RMS value of the time series
    """
    ts_frame = ts.to_frame() if isinstance(ts, pd.Series) else ts
    rms = np.zeros(ts.shape[0], dtype=float)
    factor = np.pi / (3 * np.sqrt(3))
    first_row = ts_frame.to_numpy()[0, :]
    rms[0] = factor * (np.max(first_row) - np.min(first_row))
    for n, row_values in enumerate(ts_frame.to_numpy()[1:, :], 1):
        val = factor * (np.max(row_values) - np.min(row_values))
        rms[n] = rms[n - 1] + (val - rms[n - 1]) / filter_time_constant * (
            ts_frame.index.values[n] - ts_frame.index.values[n - 1]
        )
    return pd.Series(rms, index=ts.index)
